/* MSDOS.c
   This file is part of the M2000 project
   MS-DOS dependent routines
   Copyright (C) Marcel de Kogel (m.dekogel@student.utwente.nl), 1996
   You may not use this file for commercial purposes
   Please notify me if you make any changes to this file */

#include "P2000.h"
#include "MSDOS.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/farptr.h>
#include <go32.h>
#include <dpmi.h>
#include <pc.h>
#include <dos.h>
#include <crt0.h>
#include <math.h>
#include <conio.h>

int _crt0_startup_flags = _CRT0_FLAG_NONMOVE_SBRK |
                          _CRT0_FLAG_LOCK_MEMORY  |
                          _CRT0_FLAG_DROP_EXE_SUFFIX;

#define NUM_STACKS      16
#define STACK_SIZE      8192

char *Title="M2000 MS-DOS 0.4.1";

static int *OldCharacter;
static byte *FontBuf;
word cs_alias;
static int nOldVideoMode;
__volatile__ unsigned TimerCount=0;
unsigned short DosSelector;
unsigned ReadTimerMin=47694;
static int OldTimer=0;
static int NewTimer=0;
int soundmode=255;
int joymode=1;
static int soundoff=0;

static byte regs_256x240[25]=
{
  0xE3,
  0x4F,0x3F,0x40,0x92,
  0x44,0x10,0x0D,0x3E,
  0x00,0x41,0x00,0x00,
  0x00,0x00,0x00,0x00,
  0xEA,0xAC,0xDF,0x20,
  0x40,0xE7,0x06,0xA3
};

struct font2screen_struct
{
 unsigned low;
 unsigned high;
};
static struct font2screen_struct *font2screen=NULL;

char szBitmapFile[256];

static void VGA_SetPalette (unsigned char *pal, int nColours)
{
 int i;
 __disable ();
 while ((inportb(0x3DA)&0x08)!=0); /* Wait until vertical retrace is off */
 while ((inportb(0x3DA)&0x08)==0); /* Now wait until it is on */
 outportb (0x3C8,0);               /* Start updating palette */
 for (i=0;i<nColours*3;++i)
  outportb(0x3C9,pal[i]/4);
 __enable ();
}

void VGA_Reset (void)
{
 __dpmi_regs r;
 r.x.ax=(short)nOldVideoMode;
 __dpmi_int (0x10, &r);
}

static void VGA_SetRegs (byte *regs)
{
 int i;
 __disable ();
 while (inportb(0x3DA)&8);
 while ((inportb(0x3DA)&8)==0);
 outportw (0x3C4,0x100);                        /* sequencer reset */
 outportb (0x3C2,regs[0]);                      /* misc. output reg */
 outportw (0x3C4,0x300);                        /* clear sequencer reset */
 outportw (0x3D4,((regs[0x12]&0x7F)<<8)+0x11);  /* deprotect regs 0-7 */
 for (i=0;i<24;++i)
  outportw (0x3D4,(regs[i+1]<<8)+i);
 __enable ();
}

/*** Palette ***/
static byte VGA_Palette[8*3] =
{
   0x00,0x00,0x00 , 0xFF,0x00,0x00 , 0x00,0xFF,0x00 , 0xFF,0xFF,0x00 ,
   0x00,0x00,0xFF , 0xFF,0x00,0xFF , 0x00,0xFF,0xFF , 0xFF,0xFF,0xFF
};

static int VGA_Init (void)              /* Get old mode and set mode 13h */
{
 __dpmi_regs r;
 r.x.ax=0x0F00;
 __dpmi_int (0x10,&r);
 nOldVideoMode=(int)(r.h.al & 0x7F);
 if (!P2000_Mode)
 {
  r.x.ax=0x0013;
  __dpmi_int (0x10,&r);
  VGA_SetRegs (regs_256x240);
  VGA_SetPalette (VGA_Palette,8);
 }
 else
 {
  /* Use 640x480 for P2000M emulation */
  r.x.ax=0x0012;
  __dpmi_int (0x10,&r);
  VGA_Palette[0]=VGA_Palette[1]=VGA_Palette[2]=0;
  VGA_Palette[3]=VGA_Palette[4]=VGA_Palette[5]=255;
 }
 return 1;
}

static void PutImage (void)
{
 /* The PutChar() fuctions write directly to the video memory
    Nothing to do here */
}

static __volatile__ int PausePressed=0;
static __volatile__ byte keybstatus[128];
static __volatile__ byte extkeybstatus[128];
static _go32_dpmi_seginfo newirqkeyb,oldirqkeyb;

static int keymask[256]=
{
 0x000,0x000,0x5BF,0x77F,0x0EF,0x07F,0x0DF,0x0FD,
 0x0BF,0x6BF,0x5FD,0x5DF,0x57F,0x8EF,0x5EF,0x1FE,
 0x0F7,0x4F7,0x4EF,0x47F,0x4DF,0x4FD,0x4BF,0x8BF,
 0x6FD,0x6DF,0x67F,0x7EF,0x6EF,0x0FE,0x4FB,0x1F7,
 0x1EF,0x17F,0x1DF,0x1FD,0x1BF,0x7BF,0x8FD,0x8DF,
 0x87F,0x4FE,0x9FE,0x2EF,0x1FB,0x3F7,0x3EF,0x37F,
 0x3DF,0x3FD,0x3BF,0x2BF,0x7FD,0x7DF,0x97F,0x5FE,
 0x0FB,0x2FD,0x3FE,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x5F7,0x000,0x6F7,
 0x6FB,0x6FE,0x5F7,0x8F7,0x8FB,0x8FE,0x5FB,0x7F7,
 0x7FB,0x7FE,0x2F7,0x2FB,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,

 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x2FE,0x27F,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x5FB,0x000,0x000,
 0x2DF,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x0FB,0x000,0x000,0x0FE,0x000,0x27F,0x000,0x000,
 0x2DF,0x000,0x000,0x3FB,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000
};

static int makeshot=0;
static int calloptions=0;
void keyb_interrupt (void)
{
 unsigned code;
 static int extkey;
 int tmp;
 tmp=0;
 code=inportb (0x60);           /* get scancode */
 if (code<0xE0)                 /* ignore codes >0xE0 */
 {
  if (code & 0x80)              /* key is released */
  {
   code&=0x7F;
   if (extkey)
   {
    extkeybstatus[code]=0;
    tmp=keymask[code+128];
   }
   else
   {
    keybstatus[code]=0;
    tmp=keymask[code];
   }
   if (tmp)
    KeyMap[tmp>>8]|=(~(tmp&0xFF));
  }
  else                          /* key is pressed */
  {
   if (extkey)
   {
    if (!extkeybstatus[code])
    {
     extkeybstatus[code]=1;
     PausePressed=0;
     tmp=keymask[code+128];
     switch (code)
     {
     }
    }
   }
   else
   {
    if (!keybstatus[code])
    {
     keybstatus[code]=1;
     tmp=keymask[code];
     if (code==VK_F8 || code==VK_F9)
     {
      if (PausePressed)
       PausePressed=0;
      else
      {
       if (code==VK_F8)
        PausePressed=2;
       else
        PausePressed=1;
      }
     }
     else
      PausePressed=0;
     switch (code)
     {
      case VK_Escape:
      case VK_F10:
       CPURunning=0;
       break;
#ifdef DEBUG
      case VK_F4:
       Trace=!Trace;
       break;
#endif
      case VK_F7:
       makeshot=1;
       break;
      case VK_F5:
       soundoff=(!soundoff);
       break;
      case VK_F11:
       SB_DecreaseVolume ();
       break;
      case VK_F12:
       SB_IncreaseVolume ();
       break;
      case VK_F6:
       calloptions=1;
       break;
     }
    }
   }
   if (tmp)
    KeyMap[tmp>>8]&=(tmp&0xFF);
  }
  extkey=0;
 }
 else
  if (code==0xE0)
   extkey=1;
 code=inportb (0x61);           /* acknowledge interrupt */
 outportb (0x61,code | 0x80);
 outportb (0x61,code);
 outportb (0x20,0x20);
}

extern void keyb_interrupt_asm (void);
static int Keyb_Init (void)
{
 _go32_dpmi_get_protected_mode_interrupt_vector (9,&oldirqkeyb);
 newirqkeyb.pm_offset=(int)keyb_interrupt_asm;
 newirqkeyb.pm_selector=_go32_my_cs();
 _go32_dpmi_set_protected_mode_interrupt_vector (9,&newirqkeyb);
 return 1;
}

static void Keyb_Reset (void)
{
 _go32_dpmi_set_protected_mode_interrupt_vector (9,&oldirqkeyb);
}

typedef struct joyposstruct
{
 unsigned x,y;
} joypos;

static int gotjoy=0;
static joypos joycentre;

static int _JoyGetPos (joypos *jp)
{
 unsigned tmp;
 tmp=JoyGetPos ();
 jp->x=(unsigned) (tmp&0xFFFF);
 if (jp->x>=10000)
  return 0;
 jp->y=(unsigned) (tmp>>16);
 if (jp->y>=10000)
  return 0;
 return 1;
}

static int Joy_Init (void)
{
 joypos jp;
 if (!_JoyGetPos (&jp))
  return 0;
 joycentre.x=jp.x;
 joycentre.y=jp.y;
 gotjoy=1;
 return 1;
}

static int Joy_Check (void)
{
 joypos jp;
 int J;
 int joystatus;
 if (!gotjoy)
  return 0;
 J=0;
 _JoyGetPos (&jp);
 joystatus=inportb (0x201);
 if (jp.x<(joycentre.x/2))
  J|=0x40;
 else
  if (jp.x>(joycentre.x*3/2))
   J|=0x80;
 if (jp.y<(joycentre.y/2))
  J|=0x10;
 else
  if (jp.y>((joycentre.y*3)/2))
   J|=0x20;
 if ((joystatus&0x10)==0)
  J|=0x01;
 if ((joystatus&0x20)==0)
  J|=0x02;
 return J;
}

/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
 VGA_Reset ();
 if(Verbose) printf("\n\nShutting down...\n");
 if (soundmode==2) SB_Reset ();
 RestoreTimer ();
 Keyb_Reset ();
 ExitStacks ();
 if (font2screen) free (font2screen);
 if (FontBuf) free (FontBuf);
 if (OldCharacter) free (OldCharacter);
}

static int NextBitmapFile ()
{
 char *p;
 p=szBitmapFile+strlen(szBitmapFile)-5;
 if (*p=='9')
 {
  *p='0';
  --p;
  if (*p=='9')
  {
   *p='0';
   --p;
   (*p)++;
  }
  else
  {
   (*p)++;
   if (*p=='0')
    return 0;
  }
 }
 else
  (*p)++;
 return 1;
}

/** InitMachine **********************************************/
/** Allocate resources needed by MS-DOS-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
 int c,i,j;
 FILE *bitmapfile;
 cs_alias=_my_ds();
 DosSelector=_dos_ds;
 if (Verbose)
  printf ("Allocating stack space... ");
 i=InitStacks (NUM_STACKS,STACK_SIZE);
 if (!i)
 {
  if (Verbose) puts ("FAILED");
  return 0;
 }
 if (Verbose) printf ("OK\nAllocating cache buffer... ");
 if (P2000_Mode) i=80*24;
 else i=40*24;
 OldCharacter=malloc (i*sizeof(int));
 if (!OldCharacter)
 {
  if (Verbose) puts ("FAILED");
  return(0);
 }
 memset (OldCharacter,-1,i*sizeof(int));
 if (!P2000_Mode)
 {
  if (Verbose)
   printf ("OK\nAllocating conversion buffer... ");
  font2screen=malloc(8*8*64*sizeof(struct font2screen_struct));
  if (!font2screen)
  {
   if (Verbose) puts ("FAILED");
   return 0;
  }
  if (Verbose)
   printf ("OK\nInitialising conversion buffer... ");
  for (i=0;i<8;++i)
   for (j=0;j<8;++j)
    for (c=0;c<64;++c)
    {
     unsigned a;
     a=0;
#ifdef LSB_FIRST
     if (c&0x08) a|=i; else a|=j;
     if (c&0x04) a|=(i<<8); else a|=(j<<8);
     if (c&0x02) a|=(i<<16); else a|=(j<<16);
     if (c&0x01) a|=(i<<24); else a|=(j<<24);
#else
     if (c&0x01) a|=i; else a|=j;
     if (c&0x02) a|=(i<<8); else a|=(j<<8);
     if (c&0x04) a|=(i<<16); else a|=(j<<16);
     if (c&0x08) a|=(i<<24); else a|=(j<<24);
#endif
     font2screen[i*8*64+j*64+c].high=a;
     a=0;
#ifdef LSB_FIRST
     if (c&0x20) a|=i; else a|=j;
     if (c&0x10) a|=(i<<8); else a|=(j<<8);
#else
     if (c&0x10) a|=i; else a|=j;
     if (c&0x20) a|=(i<<8); else a|=(j<<8);
#endif
     font2screen[i*8*64+j*64+c].low=a;
    }
 }
 if (Verbose) puts ("OK");
 if (joymode)
 {
  if (Verbose) printf ("Detecting joystick... ");
  i=Joy_Init ();
  if (Verbose) puts ((i)? "Found":"Not found");
 }
 strcpy (szBitmapFile,"M2000.BMP");
 while ((bitmapfile=fopen(szBitmapFile,"rb"))!=NULL)
 {
  fclose (bitmapfile);
  if (!NextBitmapFile())
   break;
 }
 if (Verbose)
  printf ("Next screenshot will be %s\n",szBitmapFile);
 if (soundmode)
 {
  if (Verbose) printf ("Detecting sound card... ");
  if (soundmode==2 || soundmode==255)
   if (!SB_Init())
   {
    if (soundmode==2) soundmode=0;
   }
   else
    soundmode=2;
  if (Verbose)
   if (!soundmode) printf ("Not found\n");
  else
   printf ("%s found\n",(soundmode!=2)? "PC Speaker":"SoundBlaster");
 }
 if (Verbose) printf ("Initialising timer...\n");
 ReadTimerMin=1192380/IFreq;
 StartTimer ();
 OldTimer=ReadTimer ();
 if (Verbose) printf ("Initialising keyboard...\n");
 Keyb_Init ();
 if (Verbose)
 {
  fprintf (stderr,"Press space to run rom code...\n");
  while (!keybstatus[VK_Space]);
 }
 VGA_Init ();
 return(1);
}

int LoadFont (char *filename)
{
 FILE *F;
 int i,c;
 char *TempBuf;
 if (Verbose) printf("Loading font %s...\n",filename);
 if (Verbose) printf ("  Allocating memory... ");
 if (!FontBuf)
 {
  FontBuf=malloc (2240);
  if (!FontBuf)
  {
   if (Verbose) puts ("FAILED");
   return 0;
  }
 }
 TempBuf=malloc (2240);
 if (!TempBuf)
 {
  if (Verbose) puts ("FAILED");
  return 0;
 }
 if (Verbose) puts ("OK");
 if (Verbose) printf ("  Opening... ");
 i=0;
 F=fopen(filename,"rb");
 if (F)
 {
  printf ("Reading... ");
  if (fread(TempBuf,2240,1,F)) i=1;
  fclose(F);
 }
 if(Verbose) puts ((i)? "OK":"FAILED");
 if(!i) return 0;
 memcpy (FontBuf,TempBuf,2240);
 free (TempBuf);
 if (Verbose) printf ("  Converting... ");
 if (P2000_Mode)
 {
  /* Stretch characters to 8x10 */
  for (i=0;i<96*10;++i)
   FontBuf[i]<<=1;
  for (i=96*10;i<(96+128)*10;++i)
  {
   c=FontBuf[i];
   FontBuf[i]=((c&7)<<1) | (c&1) | ((c&0x38)<<2) | (c&0x10);
  }
 }
 else
  /* Clear the upper two bits of each character */
  for (i=0;i<(96+128)*10;++i)
   FontBuf[i]&=0x3F;
 if (Verbose) puts ("OK");
 return 1;
}

/** Keyboard *************************************************/
/** Check for keyboard events, parse them, and modify       **/
/** keyboard controller status                              **/
/*************************************************************/
void Keyboard(void)
{
 int i;
 if (PausePressed)
 {
  i=0;
  if (PausePressed==2)          /* Blank screen */
  {
   i=1;
   inportb (0x3BA);
   inportb (0x3DA);
   outportb (0x3C0,0);
  }
  while (PausePressed);
  if (i)
  {
   inportb (0x3BA);
   inportb (0x3DA);
   outportb (0x3C0,0x20);
  }
  OldTimer=ReadTimer ();
 }
 if (makeshot)
 {
  /* Copy screen contents to buffer and write bitmap file */
  byte *p;
  sound (500);
  i=(P2000_Mode)? (640*480/8):(256*240);
  p=malloc (i);
  if (p)
  {
   dosmemget (0xA0000,i,p);
   if (!P2000_Mode)
    WriteBitmap (szBitmapFile,8,8,256,240,240,p+8,VGA_Palette);
   else
    WriteBitmap (szBitmapFile,1,2,640,640,480,p,VGA_Palette);
   free (p);
   NextBitmapFile ();
  }
  nosound ();
  makeshot=0;
 }
 if (calloptions)
 {
  calloptions=0;
  /* switch to text mode */
  VGA_Reset ();
  /* switch to default keyboard handler */
  Keyb_Reset ();
  /* call the options dialogue function */
  OptionsDialogue ();
  /* switch back to our own keyboard handler */
  Keyb_Init ();
  /* switch back to graphics mode */
  VGA_Init ();
  /* put the image back on the screen */
  memset (OldCharacter,-1,((P2000_Mode)? 80:40)*24*sizeof(int));
  RefreshScreen ();
 }
 /* Keyboard is checked in the interrupt routine
    Only check for joystick events here */
 i=Joy_Check ();
 if (i&0x40)
  KeyMap[0]&=0xFE;
 else if (!keybstatus[VK_Ctrl] && !extkeybstatus[VK_Left])
  KeyMap[0]|=0x01;
 if (i&0x80)
  KeyMap[2]&=0x7F;
 else if (!extkeybstatus[VK_Ctrl] && !extkeybstatus[VK_Right])
  KeyMap[2]|=0x80;
 if (i&0x10)
  KeyMap[0]&=0xFB;
 else if (!keybstatus[VK_Alt] && !extkeybstatus[VK_Up])
  KeyMap[0]|=0x04;
 if (i&0x20)
  KeyMap[2]&=0xDF;
 else if (!extkeybstatus[VK_Alt] && !extkeybstatus[VK_Down])
  KeyMap[2]|=0x20;
 if (i&0x0F)
  KeyMap[2]&=0xFD;
 else if (!keybstatus[VK_Space])
  KeyMap[2]|=0x02;
}

void FlushSound (void)
{
 if (soundmode==2)
  SB_FlushSound ();     /* Syncs to the SoundBlaster interrupt */
 else
 {
  if (Sync)
  {
   static int too_slow=0;
   do
    NewTimer=ReadTimer ();
   while ((NewTimer-OldTimer)<0);
   OldTimer+=ReadTimerMin;
   if ((OldTimer-NewTimer)<0)
   {
    if (++too_slow>10)
    {
     OldTimer=ReadTimer();
     too_slow=0;
    }
   }
   else
    too_slow=0;
  }
 }
}

void Sound (int toggle)
{
 int tmp;
 if (!soundmode || soundoff)
  return;
 if (soundmode==2)
 {
  SB_Sound (toggle);
  return;
 }
 tmp=inportb(0x61)&0xFC;
 if (toggle) tmp|=2;
 outportb(0x61,tmp);
}

void Pause (int ms)
{
 int i,j;
 j=ReadTimer();
 i=j+ms*1192;
 while ((j-i)<0) j=ReadTimer();
 OldTimer=j;
}

static inline void PutChar_M (int x,int y,int c,int eor,int ul)
{
 int K;
 unsigned p;
 K=c+(eor<<8)+(ul<<16);
 if (K==OldCharacter[y*80+x]) return;
 OldCharacter[y*80+x]=K;
 p=0xA0000+y*1600+x;
 _farsetsel (DosSelector);
 c*=10;
 if (eor)
 {
  K=FontBuf[c];
  _farnspokeb (p,~K);
  _farnspokeb (p+80,~K);
  K=FontBuf[c+1];
  _farnspokeb (p+160,~K);
  _farnspokeb (p+80+160,~K);
  K=FontBuf[c+2];
  _farnspokeb (p+320,~K);
  _farnspokeb (p+80+320,~K);
  K=FontBuf[c+3];
  _farnspokeb (p+480,~K);
  _farnspokeb (p+80+480,~K);
  K=FontBuf[c+4];
  _farnspokeb (p+640,~K);
  _farnspokeb (p+80+640,~K);
  K=FontBuf[c+5];
  _farnspokeb (p+800,~K);
  _farnspokeb (p+80+800,~K);
  K=FontBuf[c+6];
  _farnspokeb (p+960,~K);
  _farnspokeb (p+80+960,~K);
  K=FontBuf[c+7];
  _farnspokeb (p+1120,~K);
  _farnspokeb (p+80+1120,~K);
  K=FontBuf[c+8];
  _farnspokeb (p+1280,~K);
  _farnspokeb (p+80+1280,~K);
  if (ul) K=255; else K=FontBuf[c+9];
  _farnspokeb (p+1440,~K);
  _farnspokeb (p+80+1440,~K);
 }
 else
 {
  K=FontBuf[c];
  _farnspokeb (p,K);
  _farnspokeb (p+80,K);
  K=FontBuf[c+1];
  _farnspokeb (p+160,K);
  _farnspokeb (p+80+160,K);
  K=FontBuf[c+2];
  _farnspokeb (p+320,K);
  _farnspokeb (p+80+320,K);
  K=FontBuf[c+3];
  _farnspokeb (p+480,K);
  _farnspokeb (p+80+480,K);
  K=FontBuf[c+4];
  _farnspokeb (p+640,K);
  _farnspokeb (p+80+640,K);
  K=FontBuf[c+5];
  _farnspokeb (p+800,K);
  _farnspokeb (p+80+800,K);
  K=FontBuf[c+6];
  _farnspokeb (p+960,K);
  _farnspokeb (p+80+960,K);
  K=FontBuf[c+7];
  _farnspokeb (p+1120,K);
  _farnspokeb (p+80+1120,K);
  K=FontBuf[c+8];
  _farnspokeb (p+1280,K);
  _farnspokeb (p+80+1280,K);
  if (ul) K=255; else K=FontBuf[c+9];
  _farnspokeb (p+1440,K);
  _farnspokeb (p+80+1440,K);
 }
}

static inline void PutChar_T (int x,int y,int c,int fg,int bg,int si)
{
 int K;
 unsigned p;
 struct font2screen_struct *s;
 K=c+(fg<<8)+(bg<<16)+(si<<24);
 if (K==OldCharacter[y*40+x]) return;
 OldCharacter[y*40+x]=K;
 s=font2screen+fg*8*64+bg*64;
 p=0xA0000+y*256*10+x*6+8;
 _farsetsel (DosSelector);
 c*=10;
 if (!si)
 {
  K=FontBuf[c];
  _farnspokew(p,s[K].low);
  _farnspokel(p+2,s[K].high);
  K=FontBuf[c+1];
  _farnspokew(p+256,s[K].low);
  _farnspokel(p+2+256,s[K].high);
  K=FontBuf[c+2];
  _farnspokew(p+512,s[K].low);
  _farnspokel(p+2+512,s[K].high);
  K=FontBuf[c+3];
  _farnspokew(p+768,s[K].low);
  _farnspokel(p+2+768,s[K].high);
  K=FontBuf[c+4];
  _farnspokew(p+1024,s[K].low);
  _farnspokel(p+2+1024,s[K].high);
  K=FontBuf[c+5];
  _farnspokew(p+1280,s[K].low);
  _farnspokel(p+2+1280,s[K].high);
  K=FontBuf[c+6];
  _farnspokew(p+1536,s[K].low);
  _farnspokel(p+2+1536,s[K].high);
  K=FontBuf[c+7];
  _farnspokew(p+1792,s[K].low);
  _farnspokel(p+2+1792,s[K].high);
  K=FontBuf[c+8];
  _farnspokew(p+2048,s[K].low);
  _farnspokel(p+2+2048,s[K].high);
  K=FontBuf[c+9];
  _farnspokew(p+2304,s[K].low);
  _farnspokel(p+2+2304,s[K].high);
 }
 else
 {
  if (si==2)
   c+=5;
  K=FontBuf[c];
  _farnspokew(p,s[K].low);
  _farnspokew(p+256,s[K].low);
  _farnspokel(p+2,s[K].high);
  _farnspokel(p+2+256,s[K].high);
  K=FontBuf[c+1];
  _farnspokew(p+512,s[K].low);
  _farnspokew(p+768,s[K].low);
  _farnspokel(p+2+512,s[K].high);
  _farnspokel(p+2+768,s[K].high);
  K=FontBuf[c+2];
  _farnspokew(p+1024,s[K].low);
  _farnspokew(p+1280,s[K].low);
  _farnspokel(p+2+1024,s[K].high);
  _farnspokel(p+2+1280,s[K].high);
  K=FontBuf[c+3];
  _farnspokew(p+1536,s[K].low);
  _farnspokew(p+1792,s[K].low);
  _farnspokel(p+2+1536,s[K].high);
  _farnspokel(p+2+1792,s[K].high);
  K=FontBuf[c+4];
  _farnspokew(p+2048,s[K].low);
  _farnspokew(p+2304,s[K].low);
  _farnspokel(p+2+2048,s[K].high);
  _farnspokel(p+2+2304,s[K].high);
 }
}

/** Part of the code common for Unix/X and MSDOS drivers *****/
#include "Common.h"
