/* SVGA.c
   This file is part of the M2000 project
   Linux/SVGALib dependent routines
   Copyright (C) Marcel de Kogel (m.dekogel@student.utwente.nl), 1996
   You may not use this file for commercial purposes
   Please notify me if you make any changes to this file */

#include "P2000.h"
#include "Unix.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <termio.h>
#include <vga.h>
#include <linux/keyboard.h>
#include <vgakeyboard.h>
#include <asm/io.h>

#ifndef SCANCODE_F11
#define SCANCODE_F11	87
#endif
#ifndef SCANCODE_F12
#define SCANCODE_F12	88
#endif

char *Title="M2000 Linux/SVGALib 0.4.1";

byte *DisplayBuf;
static int *OldCharacter;
static byte *FontBuf;
unsigned ReadTimerMin;
static int OldTimer;
static int NewTimer;
#ifdef SOUND
int soundmode=255;
static int soundoff=0;
#endif
#ifdef JOYSTICK
int joymode=1;
#endif
int chipset=0;
static struct termios termold;

struct font2screen_struct
{
 unsigned low;
 unsigned high;
};
static struct font2screen_struct *font2screen=NULL;

char szBitmapFile[256];

/*** Palette ***/
static byte Pal[8*3] =
{
   0x00,0x00,0x00 , 0xFF,0x00,0x00 , 0x00,0xFF,0x00 , 0xFF,0xFF,0x00 ,
   0x00,0x00,0xFF , 0xFF,0x00,0xFF , 0x00,0xFF,0xFF , 0xFF,0xFF,0xFF
};

static inline volatile void outportw (word _port,word _data)
{
 outw (_data,_port);
}

static void PutImage (void)            /* Copy DisplayBuf to screen */
{
 if (!P2000_Mode)
 {
  char *p,*z;
  int i,j;
  p=vga_getgraphmem()+32/4;
  z=DisplayBuf;
  for (i=0;i<240;++i,p+=320/4,z+=256)
  {
   outportw (0x3C4,0x102);
   for (j=0;j<64;++j)
    p[j]=z[j*4+0];
   outportw (0x3C4,0x202);
   for (j=0;j<64;++j)
    p[j]=z[j*4+1];
   outportw (0x3C4,0x402);
   for (j=0;j<64;++j)
    p[j]=z[j*4+2];
   outportw (0x3C4,0x802);
   for (j=0;j<64;++j)
    p[j]=z[j*4+3];
  }
 }
 else
 {
  byte gcregs[9]= {0,0,0,0,0,0,5,15,255};
  int i;
  /* SVGALib messes with these */
  for (i=0;i<9;++i)
   outportw (0x3CE,i+(gcregs[i]<<8));
  memcpy (vga_getgraphmem(),DisplayBuf,640*480/8);
 }
}

static int PausePressed=0;
static int makeshot=0;
static int calloptions=0;
static byte keybstatus[NR_KEYS];

static int keymask[NR_KEYS]=
{
 0x000,0x000,0x5BF,0x77F,0x0EF,0x07F,0x0DF,0x0FD,
 0x0BF,0x6BF,0x5FD,0x5DF,0x57F,0x8EF,0x5EF,0x1FE,
 0x0F7,0x4F7,0x4EF,0x47F,0x4DF,0x4FD,0x4BF,0x8BF,
 0x6FD,0x6DF,0x67F,0x7EF,0x6EF,0x0FE,0x4FB,0x1F7,
 0x1EF,0x17F,0x1DF,0x1FD,0x1BF,0x7BF,0x8FD,0x8DF,
 0x87F,0x4FE,0x9FE,0x2EF,0x1FB,0x3F7,0x3EF,0x37F,
 0x3DF,0x3FD,0x3BF,0x2BF,0x7FD,0x7DF,0x97F,0x5FE,
 0x0FB,0x2FD,0x3FE,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x5F7,0x000,0x6F7,
 0x6FB,0x6FE,0x5F7,0x8F7,0x8FB,0x8FE,0x5FB,0x7F7,
 0x7FB,0x7FE,0x2F7,0x2FB,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x2FE,0x27F,0x5FB,0x000,0x2DF,0x000,0x000,0x0FB,
 0x000,0x0FE,0x27F,0x000,0x2DF,0x000,0x000,0x3FB,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
 0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000
};

static void keyb_handler (int code,int newstatus)
{
 int tmp;
 if (code<0 || code>=NR_KEYS)
  return;
 if (newstatus) newstatus=1;
 if (keybstatus[code]!=newstatus)
 {
  if (!newstatus)
  {
   keybstatus[code]=0;
   tmp=keymask[code];
   if (tmp)
    KeyMap[tmp>>8]|=(~(tmp&0xFF));
  }
  else
  {
   keybstatus[code]=1;
   tmp=keymask[code];
   if (code==SCANCODE_F8 || code==SCANCODE_F9)
   {
    if (PausePressed)
     PausePressed=0;
    else
    {
     if (code==SCANCODE_F8)
      PausePressed=2;
     else
      PausePressed=1;
    }
   }
   else
    PausePressed=0;
   switch (code)
   {
    case SCANCODE_ESCAPE:
    case SCANCODE_F10:
     CPURunning=0;
     break;
#ifdef DEBUG
    case SCANCODE_F4:
     Trace=!Trace;
     break;
#endif
    case SCANCODE_F7:
     makeshot=1;
     break;
#ifdef SOUND
    case SCANCODE_F5:
     soundoff=(!soundoff);
     break;
    case SCANCODE_F11:
     DecreaseSoundVolume ();
     break;
    case SCANCODE_F12:
     IncreaseSoundVolume ();
     break;
#endif
    case SCANCODE_F6:
     calloptions=1;
     break;
   }
   if (tmp)
    KeyMap[tmp>>8]&=(tmp&0xFF);
  }
 }
}

/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine()         **/
/*************************************************************/
void TrashMachine(void)
{
 if (Verbose) printf("\n\nShutting down...\n");
#ifdef SOUND
 TrashSound ();
#endif
#ifdef JOYSTICK
 TrashJoystick ();
#endif
 keyboard_close ();
 vga_setmode (TEXT);
 tcsetattr (0,0,&termold);
 if (font2screen) free (font2screen);
 if (DisplayBuf) free (DisplayBuf);
 if (FontBuf) free (FontBuf);
 if (OldCharacter) free (OldCharacter);
}

static int NextBitmapFile ()
{
 char *p;
 p=szBitmapFile+strlen(szBitmapFile)-5;
 if (*p=='9')
 {
  *p='0';
  --p;
  if (*p=='9')
  {
   *p='0';
   --p;
   (*p)++;
  }
  else
  {
   (*p)++;
   if (*p=='0')
    return 0;
  }
 }
 else
  (*p)++;
 return 1;
}

/** InitMachine ************************************************/
/** Allocate resources needed by Linux/SVGALib-dependent code **/
/***************************************************************/
int InitMachine(void)
{
 int c,i,j;
 FILE *bitmapfile;
 if (Verbose)
  printf ("Initialising Linux/SVGALib drivers:\n");
 tcgetattr (0,&termold);
 if (!chipset) vga_setchipset (VGA);
 vga_init ();
 if (Verbose) printf ("  Setting VGA mode... ");
 if (!P2000_Mode)
  i=G320x240x256;
 else
  i=G640x480x16;
 i=vga_setmode (i);
 if (i) { if (Verbose) printf ("FAILED\n"); return 0; }
 if (!P2000_Mode)
 {
  for (i=0;i<8;++i)
   vga_setpalette (i,Pal[i*3+0]>>2,Pal[i*3+1]>>2,Pal[i*3+2]>>2);
 }
 else
 {
  Pal[0]=Pal[1]=Pal[2]=0;
  Pal[3]=Pal[4]=Pal[5]=0;
 }
 if(Verbose)
  printf("OK\n  Allocating screen buffer... ");
 if (P2000_Mode) i=640*480/8;
 else i=256*240+sizeof(unsigned)-4;
 DisplayBuf=(byte *)malloc(i);
 if (!DisplayBuf)
 {
  if (Verbose) puts ("FAILED");
  return(0);
 }
 memset (DisplayBuf,0,i);
 if (Verbose) printf ("OK\n  Allocating cache buffer... ");
 if (P2000_Mode) i=80*24;
 else i=40*24;
 OldCharacter=malloc (i*sizeof(int));
 if (!OldCharacter)
 {
  if (Verbose) puts ("FAILED");
  return(0);
 }
 memset (OldCharacter,-1,i*sizeof(int));
 if (!P2000_Mode)
 {
  if (Verbose)
   printf ("OK\n  Allocating conversion buffer... ");
  font2screen=malloc(8*8*64*sizeof(struct font2screen_struct));
  if (!font2screen)
  {
   if (Verbose) puts ("FAILED");
   return 0;
  }
  if (Verbose)
   printf ("OK\n  Initialising conversion buffer... ");
  for (i=0;i<8;++i)
   for (j=0;j<8;++j)
    for (c=0;c<64;++c)
    {
     unsigned a;
     a=0;
#ifdef LSB_FIRST
     if (c&0x08) a|=i; else a|=j;
     if (c&0x04) a|=(i<<8); else a|=(j<<8);
     if (c&0x02) a|=(i<<16); else a|=(j<<16);
     if (c&0x01) a|=(i<<24); else a|=(j<<24);
#else
     if (c&0x01) a|=i; else a|=j;
     if (c&0x02) a|=(i<<8); else a|=(j<<8);
     if (c&0x04) a|=(i<<16); else a|=(j<<16);
     if (c&0x08) a|=(i<<24); else a|=(j<<24);
#endif
     font2screen[i*8*64+j*64+c].high=a;
     a=0;
#ifdef LSB_FIRST
     if (c&0x20) a|=i; else a|=j;
     if (c&0x10) a|=(i<<8); else a|=(j<<8);
#else
     if (c&0x10) a|=i; else a|=j;
     if (c&0x20) a|=(i<<8); else a|=(j<<8);
#endif
     font2screen[i*8*64+j*64+c].low=a;
    }
 }
 if (Verbose) puts ("OK");
#ifdef JOYSTICK
 InitJoystick (joymode);
#endif
 strcpy (szBitmapFile,"M2000.bmp");
 while ((bitmapfile=fopen(szBitmapFile,"rb"))!=NULL)
 {
  fclose (bitmapfile);
  if (!NextBitmapFile())
   break;
 }
 if (Verbose)
  printf ("  Next screenshot will be %s\n",szBitmapFile);
#ifdef SOUND
  InitSound (soundmode);
#endif
 if (Verbose) printf ("  Initialising timer...\n");
 ReadTimerMin=1000000/IFreq;
 OldTimer=ReadTimer ();
 if (Verbose) printf ("  Initialising keyboard...");
 if (keyboard_init()) { if (Verbose) printf ("FAILED\n"); return 0; }
 keyboard_seteventhandler (keyb_handler);
 if (Verbose) printf ("OK\n");
 return 1;
}

int LoadFont (char *filename)
{
 FILE *F;
 int i,c;
 char *TempBuf;
 if (Verbose) printf("Loading font %s...\n",filename);
 if (Verbose) printf ("  Allocating memory... ");
 if (!FontBuf)
 {
  FontBuf=malloc (2240);
  if (!FontBuf)
  {
   if (Verbose) puts ("FAILED");
   return 0;
  }
 }
 TempBuf=malloc (2240);
 if (!TempBuf)
 {
  if (Verbose) puts ("FAILED");
  return 0;
 }
 if (Verbose) puts ("OK");
 if (Verbose) printf ("  Opening... ");
 i=0;
 F=fopen(filename,"rb");
 if (F)
 {
  if (Verbose) printf ("Reading... ");
  if (fread(TempBuf,2240,1,F)) i=1;
  fclose(F);
 }
 if(Verbose) puts ((i)? "OK":"FAILED");
 if(!i) return 0;
 memcpy (FontBuf,TempBuf,2240);
 free (TempBuf);
 if (Verbose) printf ("  Converting... ");
 if (P2000_Mode)
 {
  /* Stretch characters to 8x10 */
  for (i=0;i<96*10;++i)
   FontBuf[i]<<=1;
  for (i=96*10;i<(96+128)*10;++i)
  {
   c=FontBuf[i];
   FontBuf[i]=((c&7)<<1) | (c&1) | ((c&0x38)<<2) | (c&0x10);
  }
 }
 else
  /* Clear the upper two bits of each character */
  for (i=0;i<(96+128)*10;++i)
   FontBuf[i]&=0x3F;
 if (Verbose) puts ("OK");
 return 1;
}

/** Keyboard *************************************************/
/** Check for keyboard events, parse them, and modify       **/
/** keyboard controller status                              **/
/*************************************************************/
void Keyboard(void)
{
 int i;
 char *p;
 keyboard_update ();
 if (PausePressed)
 {
  if (PausePressed==2)          /* Blank screen */
  {
   outportw (0x3C4,0xF02);
   p=vga_getgraphmem ();
   for (i=0;i<320*200;++i) *p++=0;
  }
  while (PausePressed) keyboard_update ();
  OldTimer=ReadTimer ();
 }
 if (makeshot)
 {
  if (!P2000_Mode)
   WriteBitmap (szBitmapFile,8,8,256,240,240,DisplayBuf+8,Pal);
  else
   WriteBitmap (szBitmapFile,1,2,640,640,480,DisplayBuf,Pal);
  NextBitmapFile ();
  makeshot=0;
 }
 if (calloptions)
 {
  struct termios term;
  calloptions=0;
  vga_flip ();
  keyboard_close ();
  tcgetattr (0,&term);
  tcsetattr (0,0,&termold);
  OptionsDialogue ();
  tcsetattr (0,0,&term);
  keyboard_init ();
  keyboard_seteventhandler (keyb_handler);
  vga_flip ();
 }
#ifdef JOYSTICK
 i=ReadJoystick ();
 if (i&0x40)
  KeyMap[0]&=0xFE;
 else if (!keybstatus[SCANCODE_LEFTCONTROL] &&
          !keybstatus[SCANCODE_CURSORBLOCKLEFT])
  KeyMap[0]|=0x01;
 if (i&0x80)
  KeyMap[2]&=0x7F;
 else if (!keybstatus[SCANCODE_RIGHTCONTROL] &&
          !keybstatus[SCANCODE_CURSORBLOCKRIGHT])
  KeyMap[2]|=0x80;
 if (i&0x10)
  KeyMap[0]&=0xFB;
 else if (!keybstatus[SCANCODE_LEFTALT] &&
          !keybstatus[SCANCODE_CURSORBLOCKUP])
  KeyMap[0]|=0x04;
 if (i&0x20)
  KeyMap[2]&=0xDF;
 else if (!keybstatus[SCANCODE_RIGHTALT] &&
          !keybstatus[SCANCODE_CURSORBLOCKDOWN])
  KeyMap[2]|=0x20;
 if (i&0x0F)
  KeyMap[2]&=0xFD;
 else if (!keybstatus[SCANCODE_SPACE])
  KeyMap[2]|=0x02;
#endif
}

void FlushSound (void)
{
#ifdef SOUND
 if (!Sound_FlushSound())
#endif
 {
  if (Sync)
  {
   static int too_slow=0;
   do
    NewTimer=ReadTimer ();
   while ((NewTimer-OldTimer)<0);
   OldTimer+=ReadTimerMin;
   if ((OldTimer-NewTimer)<0)
   {
    if (++too_slow>10)
    {
     OldTimer=ReadTimer();
     too_slow=0;
    }
   }
   else
    too_slow=0;
  }
 }
}

void Sound (int toggle)
{
#ifdef SOUND
 if (!soundoff) WriteSound (toggle);
#endif
}

void Pause (int ms)
{
 int i,j;
 j=ReadTimer();
 i=j+ms*1000;
 while ((j-i)<0) j=ReadTimer();
 OldTimer=j;
}

static void PutChar_M (int x,int y,int c,int eor,int ul)
{
 int K;
 byte *p;
 K=c+(eor<<8)+(ul<<16);
 if (K==OldCharacter[y*80+x]) return;
 OldCharacter[y*80+x]=K;
 p=DisplayBuf+y*1600+x;
 c*=10;
 if (eor)
 {
  K=FontBuf[c];
  p[0]=p[80]=~K;
  K=FontBuf[c+1];
  p[160]=p[80+160]=~K;
  K=FontBuf[c+2];
  p[320]=p[80+320]=~K;
  K=FontBuf[c+3];
  p[480]=p[80+480]=~K;
  K=FontBuf[c+4];
  p[640]=p[80+640]=~K;
  K=FontBuf[c+5];
  p[800]=p[80+800]=~K;
  K=FontBuf[c+6];
  p[960]=p[80+960]=~K;
  K=FontBuf[c+7];
  p[1120]=p[80+1120]=~K;
  K=FontBuf[c+8];
  p[1280]=p[80+1280]=~K;
  if (ul) K=255; else K=FontBuf[c+9];
  p[1440]=p[80+1440]=~K;
 }
 else
 {
  K=FontBuf[c];
  p[0]=p[80]=K;
  K=FontBuf[c+1];
  p[160]=p[80+160]=K;
  K=FontBuf[c+2];
  p[320]=p[80+320]=K;
  K=FontBuf[c+3];
  p[480]=p[80+480]=K;
  K=FontBuf[c+4];
  p[640]=p[80+640]=K;
  K=FontBuf[c+5];
  p[800]=p[80+800]=K;
  K=FontBuf[c+6];
  p[960]=p[80+960]=K;
  K=FontBuf[c+7];
  p[1120]=p[80+1120]=K;
  K=FontBuf[c+8];
  p[1280]=p[80+1280]=K;
  if (ul) K=255; else K=FontBuf[c+9];
  p[1440]=p[80+1440]=K;
 }
}

static inline void PutChar_T (int x,int y,int c,int fg,int bg,int si)
{
 int K;
 byte *p;
 struct font2screen_struct *s;
 K=c+(fg<<8)+(bg<<16)+(si<<24);
 if (K==OldCharacter[y*40+x]) return;
 OldCharacter[y*40+x]=K;
 s=font2screen+fg*8*64+bg*64;
 p=DisplayBuf+y*256*10+x*6+8;
 c*=10;
 if (!si)
 {
  K=FontBuf[c];
  *(unsigned short*)p=(unsigned short)s[K].low;
  *(unsigned*)(p+2)=s[K].high;
  K=FontBuf[c+1];
  *(unsigned short*)(p+256)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+256)=s[K].high;
  K=FontBuf[c+2];
  *(unsigned short*)(p+512)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+512)=s[K].high;
  K=FontBuf[c+3];
  *(unsigned short*)(p+768)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+768)=s[K].high;
  K=FontBuf[c+4];
  *(unsigned short*)(p+1024)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+1024)=s[K].high;
  K=FontBuf[c+5];
  *(unsigned short*)(p+1280)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+1280)=s[K].high;
  K=FontBuf[c+6];
  *(unsigned short*)(p+1536)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+1536)=s[K].high;
  K=FontBuf[c+7];
  *(unsigned short*)(p+1792)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+1792)=s[K].high;
  K=FontBuf[c+8];
  *(unsigned short*)(p+2048)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+2048)=s[K].high;
  K=FontBuf[c+9];
  *(unsigned short*)(p+2304)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+2304)=s[K].high;
 }
 else
 {
  if (si==2)
   c+=5;
  K=FontBuf[c];
  *(unsigned short*)p=*(unsigned short*)(p+256)=(unsigned short)s[K].low;
  *(unsigned*)(p+2)=*(unsigned*)(p+2+256)=s[K].high;
  K=FontBuf[c+1];
  *(unsigned short*)(p+512)=*(unsigned short*)(p+256+512)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+512)=*(unsigned*)(p+2+256+512)=s[K].high;
  K=FontBuf[c+2];
  *(unsigned short*)(p+1024)=*(unsigned short*)(p+256+1024)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+1024)=*(unsigned*)(p+2+256+1024)=s[K].high;
  K=FontBuf[c+3];
  *(unsigned short*)(p+1536)=*(unsigned short*)(p+256+1536)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+1536)=*(unsigned*)(p+2+256+1536)=s[K].high;
  K=FontBuf[c+4];
  *(unsigned short*)(p+2048)=*(unsigned short*)(p+256+2048)=(unsigned short)s[K].low;
  *(unsigned*)(p+2+2048)=*(unsigned*)(p+2+256+2048)=s[K].high;
 }
}

/** Part of the code common for Unix/X and MSDOS drivers *****/
#include "Common.h"
