/** Z80: portable Z80 emulator *******************************/
/**                                                         **/
/**                          Z80IO.h                        **/
/**                                                         **/
/** Copyright (C) Marcel de Kogel 1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/   
/**     changes to this file.                               **/
/*************************************************************/

#define INLINE_MEM      /* Inline M_RDMEM() and M_WRMEM() */

/*** Input/Output values from/to ports: *******************/
/*** These are called on IN and OUT commands and should ***/
/*** supplied by machine-dependent part of the code.    ***/
/**********************************************************/
byte DoIn(byte Port)                   FASTCALL;
void DoOut(byte Port,byte Value)       FASTCALL;

/**********************************************************/
/*** These functions are called when read or write to   ***/
/*** RAM occurs. They perform actual reads abd writes.  ***/
/**********************************************************/
#ifdef INLINE_MEM
 extern byte *ReadPage[256];
 extern byte *WritePage[256];
 INLINE unsigned M_RDMEM (dword a)
 {
  return ReadPage[a>>8][a&0xFF];
 }
 INLINE void M_WRMEM(dword a,byte v)
 {
  WritePage[a>>8][a&0xFF]=v;
 }
#else
 byte M_RDMEM (dword A)         FASTCALL;
 void M_WRMEM (dword A,byte V)  FASTCALL;
#endif

/* Used by PUSH and POP instructions */
#define M_RDMEM_FAST(A)        M_RDMEM(A)
#define M_WRMEM_FAST(A,V)      M_WRMEM(A,V)

/* Get next opcode and increment program counter */
INLINE unsigned M_RDMEM_OPCODE (void)
{
 unsigned retval;
 retval=M_RDMEM_FAST(R.PC.D);
 R.PC.W.l++;
 return retval;
}

